from __future__ import annotations

import sys
from typing import Any
from typing import Optional

import requests
from intake.catalog import Catalog
from intake.catalog.local import CatalogParser

from . import __version__
from .exceptions import AnacondaCatalogsError
from .exceptions import ClientRequiresUpdate

DEFAULT_BASE_URI = "https://anaconda.cloud/api"
API_VERSION = "2023.02.20"
USER_AGENT = f"anaconda-catalogs/{__version__}"


class AnacondaCatalog(Catalog):
    name = "anaconda_catalog"

    def __init__(self, name: str, base_uri: Optional[str] = None, **kwargs: Any):
        self.name = name
        self.base_uri = base_uri or DEFAULT_BASE_URI
        super().__init__(name=name, **kwargs)

    def _get_from_server(self) -> dict[str, Any]:
        """Load the catalog spec from the API."""
        # Note: This is split out as a separate method to enable easier mocking
        url = slash_join(self.base_uri, "catalogs", self.name)
        response = requests.get(
            url,
            headers={
                "Accept": "application/json",
                "Api-Version": API_VERSION,
                "User-Agent": USER_AGENT,
            },
        )
        # Even before trying the standard response handling
        # See if the backend server explicitly wants us to perform behavior
        # such as printing to the user or displaying a custom response
        data: dict[str, Any] = safe_get_json(response) or {}
        self._handle_response_display_message(data.get("display_message", {}) or {})

        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            raise AnacondaCatalogsError(str(e))
        return response.json()

    def _get_catalog_spec(self) -> dict[str, Any]:
        """Load the Intake-parsable spec from the service, handling nested catalogs."""
        response_data = self._get_from_server()
        catalogs = response_data.get("catalogs")
        if catalogs is None:
            # It is a single catalog
            return response_data["spec"]

        # Otherwise, it's a catalog of catalogs
        sources = {}
        for catalog in catalogs:
            cid = catalog["id"]
            name = catalog["spec"].get("name") or cid
            normalized_name = "_".join(name.lower().split())
            sources[normalized_name] = {
                "driver": "anaconda_catalog",
                "args": {"name": cid},
            }
        return {"sources": sources}

    def _load(self) -> None:
        """Populate the catalog by loading the spec from the catalogs service."""
        spec = self._get_catalog_spec()

        # Parse the catalog spec. This is the same way Intake parses YAML catalogs internally.
        context = {"root": self.name}
        result = CatalogParser(spec, context=context)

        self._entries = {}

        cfg: dict[str, Any] = result.data or {}
        for entry in cfg["data_sources"]:
            entry._catalog = self
            self._entries[entry.name] = entry

        self.metadata.update(cfg.get("metadata") or {})
        self.description: str = self.description or cfg.get("description", "")

    def _handle_response_display_message(self, display_message: dict[str, Any]) -> None:
        if message := display_message.get("message_of_the_day"):
            print(message, file=sys.stderr)
        error_code = display_message.get("error_code", None)
        error_message = display_message.get("error_message", "")
        if error_code == "client_requires_update":
            raise ClientRequiresUpdate(
                error_message
                or f"anaconda-catalogs {__version__} is out-of-date. Please upgrade to the latest version",
            )
        elif error_code == "unknown_error":
            raise AnacondaCatalogsError(
                error_message or "An unknown error has occurred"
            )


# It really shouldn't be so complicated to safely create a URL from components...
# https://codereview.stackexchange.com/a/175423
def slash_join(*args: str) -> str:
    return "/".join(arg.strip("/") for arg in args)


def safe_get_json(response: requests.Response) -> Optional[dict[str, Any]]:
    try:
        return response.json()
    except requests.exceptions.JSONDecodeError:
        return None



"""Exceptions specifically related to the Anaconda catalogs service."""

from requests import HTTPError

# TODO: Add more refined exceptions


class AnacondaCatalogsError(HTTPError):
    pass


class ClientRequiresUpdate(HTTPError):
    pass



"""anaconda.catalogs is a library allowing users to interface with Anaconda's Intake catalogs service."""
import importlib.metadata
from typing import Optional

from intake import Catalog

# __version__ is imported before the catalog subclass
# to avoid circular imports
__version__ = importlib.metadata.version("anaconda-catalogs")

from .catalog import AnacondaCatalog

__all__ = ["__version__", "open_catalog", "AnacondaCatalog"]


def open_catalog(name: str, base_uri: Optional[str] = None) -> Catalog:
    """Open an Intake catalog hosted on Anaconda Catalogs."""
    return AnacondaCatalog(name=name, base_uri=base_uri)



from __future__ import annotations

from pathlib import Path
from textwrap import dedent
from typing import TYPE_CHECKING
from typing import Any

import intake
import pytest
import yaml

import anaconda_catalogs
from anaconda_catalogs import AnacondaCatalog

if TYPE_CHECKING:
    from _pytest.monkeypatch import MonkeyPatch


@pytest.fixture()
def mocked_catalogs_service(tmp_path: Path, monkeypatch: MonkeyPatch) -> None:
    """Monkey-patch `anaconda_catalogs.AnacondaCatalog` to return test data from a mock API."""
    catalog_yaml = dedent(
        """\
        sources:
            wind_turbines:
              args:
                urlpath: "zip://*.csv::https://eerscmap.usgs.gov/uswtdb/assets/data/uswtdbCSV.zip"

                csv_kwargs:
                  assume_missing: true

              description: "US Wind Turbine Database"
              driver: csv
        """
    )

    def _get_mocked_catalog_spec(self: AnacondaCatalog) -> dict[str, Any]:
        return yaml.safe_load(catalog_yaml)

    monkeypatch.setattr(
        AnacondaCatalog,
        "_get_catalog_spec",
        _get_mocked_catalog_spec,
    )


@pytest.fixture(
    params=[
        "anaconda_catalogs.AnacondaCatalog",
        "anaconda_catalogs.catalog.AnacondaCatalog",
        "anaconda_catalog",
    ]
)
def catalog_yaml_path(request, tmp_path):
    catalog = dedent(
        f"""\
        sources:
          anaconda:
            driver: {request.param}
            args:
              name: cid/mocked
        """
    )

    catalog_path = tmp_path / "catalog.yaml"
    with catalog_path.open("wt") as f:
        f.write(catalog)

    yield catalog_path


@pytest.mark.parametrize("method", ["direct_import", "plugin"])
@pytest.mark.usefixtures("mocked_catalogs_service")
def test_open_catalog(method: str):
    # Open the catalog
    if method == "direct_import":
        cat = anaconda_catalogs.open_catalog("cid/not-a-real-catalog")
    elif method == "plugin":
        cat = intake.open_anaconda_catalog("cid/not-a-real-catalog")
    else:
        raise ValueError("Unreachable")

    assert cat.name == "cid/not-a-real-catalog"

    # There are data sources
    assert list(cat)

    # The "wind_turbine" data source exists in the catalog
    assert "wind_turbines" in cat


@pytest.mark.usefixtures("mocked_catalogs_service")
def test_open_anaconda_catalog_kwarg():
    cat = intake.open_anaconda_catalog(name="cid/mocked")
    assert cat.name == "cid/mocked"
    assert "wind_turbines" in cat


@pytest.mark.usefixtures("mocked_catalogs_service")
def test_open_catalog_yaml(catalog_yaml_path):
    cat = intake.open_catalog(str(catalog_yaml_path))

    # this is how yaml_file_cat works
    assert cat.name == catalog_yaml_path.parent.name

    assert "anaconda" in cat

    # the name is chosen by the catalog.yaml
    assert cat.anaconda.name == "anaconda"

    assert "wind_turbines" in cat.anaconda
